"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ice sheet and ice shelf'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ice sheets and ice shelves',
    'properties': [
        ('grounding_line_method', 'ENUM:grounding_line_methods', '1.1',
            'Specify the technique used for modelling the grounding line in the ice sheet-ice shelf coupling'),
        ('ice_sheet', 'bool', '1.1',
            'Are ice sheets simulated?'),
        ('ice_shelf', 'bool', '1.1',
            'Are ice shelves simulated?'),
    ]
}

#where I had calculated the SMB after I run the climate model for a year. Here we used sub-daily atmospheric output. For another model setup, we had calculated the SMB in the atmosphere model component on the atmosphere grid. Since this grid is too coarse to resolve the marginal melting zones, we are planning to couple the atmosphere model component to a higher resolved specialized SMB model. Therefore, we want to transfer the required fields through our coupler, like it is done for the atmosphere-ocean coupling. The SMB model will perform it's own IO. 

# --------------------------------------------------------------------
# SUB-PROCESS: Mass balance
# --------------------------------------------------------------------
DETAILS['mass_balance'] = {
    'description': 'Description of the surface mass balance treatment',
    'properties': [
        ('surface_mass_balance', 'str', '1.1',
            'Describe how and where the surface mass balance (SMB) is calulated. Include the temporal coupling frequeny from the atmosphere, whether or not a seperate  SMB model is used, and if so details of this model, such as its resolution'),
    ],
    'detail_sets': [
        'basal',
        'frontal'
    ],
}

DETAILS['mass_balance:basal'] = {
    'description': 'Description of basal melting',
    'properties': [
        ('bedrock', 'str', '0.1',
             'Describe the implementation of basal melting over bedrock'),
        ('ocean', 'str', '0.1',
             'Describe the implementation of basal melting over the ocean'),
    ]
}

DETAILS['mass_balance:frontal'] = {
    'description': 'Description of claving/melting from the ice shelf front',
    'properties': [
        ('calving', 'str', '0.1',
             'Describe the implementation of calving from the front of the ice shelf'),
        ('melting', 'str', '0.1',
             'Describe the implementation of melting from the front of the ice shelf'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Dynamics
# --------------------------------------------------------------------
DETAILS['dynamics'] = {
    'description': '',
    'properties': [
        ('description', 'str', '1.1',
            'General description if ice sheet and ice shelf dynamics'),
        ('approximation', 'ENUM:approximation_types', '1.N',
            'Approximation type used in modelling ice dynamics'),
        ('adaptive_timestep', 'bool', '1.1',
             'Is there an adaptive time scheme for the ice scheme?'),
        ('timestep', 'int', '1.1',
             'Timestep (in seconds) of the ice scheme. If the timestep is adaptive, then state a representative timestep.',),
#        ('coupling_wth_atmosphere', 'str', '0.1',
#             'Describe the coupling method between the ice sheet and ice shelf and atmosphe#re'),
#        ('coupling_wth_ocean', 'str', '0.1',
#             'Describe the coupling method between the ice sheet and ice shelf and the ocea#n'),
#    ],
#    'detail_sets': [
#        'numerics',
    ]
}

#DETAILS['dynamics:numerics'] = {
#    'description': 'TODO',
#    'properties': [
#        ('timestep', 'int', '1.1',
#             'Timestep (in seconds) of the ice scheme',),        
#        ('approximation', 'ENUM:approximation_types', '1.N',
#            'Approximation type used in modelling ice dynamics'),
#   ],
#}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['grounding_line_methods'] = {
    'description': 'Specify the technique used for modelling the grounding line in the ice sheet-ice shelf coupling',
    'is_open': True,
    'members': [
        ('grounding line prescribed', None),
        ('flux prescribed (Schoof)', None),
        ('fixed grid size', None),
        ('moving grid', None),
    ]
}

ENUMERATIONS['ablation_calculation_types'] = {
    'description': 'Type of scheme used to model ice sheet ablation',
    'is_open': True,
    'members': [
        ('Energy balance model', None),
        ('PDD', 'positive degree days'),
    ]
}

ENUMERATIONS['approximation_types'] = {
    'description': 'Approximation type used in modelling ice sheet dynamics',
    'is_open': True,
    'members': [
        ('SIA', None),
        ('SAA', None),
        ('full stokes', None),
    ]
}

ENUMERATIONS['ice_sheet_snow_processes'] = {
    'description': 'Processes describing snow on ice sheets',
    'is_open': True,
    'members': [
        ('snow melting', None),
        ('snow refreezing', None),
        ('blowing snow', None),
    ]
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry startospheric heterogeneous chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Gas phase chemistry attributes',
    'properties': [
        ('overview', 'str', '1.1',
            'Overview stratospheric heterogenous atmospheric chemistry'), 
        ('gas_phase_species', 'ENUM:gas_phase_species', '0.N',
            'Gas phase species included in the stratospheric heterogeneous chemistry scheme.'),
        ('aerosol_species', 'ENUM:aerosol_species', '0.N',
            'Aerosol species included in the stratospheric heterogeneous chemistry scheme.'),
        ('number_of_steady_state_species', 'int', '1.1',
             'The number of steady state species in the stratospheric heterogeneous chemistry scheme.'),
        ('sedimentation', 'bool', '1.1',
            'Is sedimentation is included in the stratospheric heterogeneous chemistry scheme or not?'),       
        ('coagulation', 'bool', '1.1',
            'Is coagulation is included in the stratospheric heterogeneous chemistry scheme or not?'),       
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['gas_phase_species'] = {
    'description': 'Gas phase species included in the stratospheric heterogeneous chemistry scheme',
    'is_open': False,
    'members': [
        ('Cly', None),
        ('Bry', None),
        ('NOy', None),
    ]
}

ENUMERATIONS['aerosol_species'] = {
    'description': 'Aerosol species included in the stratospheric heterogeneous chemistry scheme',
    'is_open': False,
    'members': [
        ('Sulphate', None),
        ('Polar stratospheric ice', None),
        ('NAT (Nitric acid trihydrate)', None),
        ('NAD (Nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particule))', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: security.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps security related functions.

.. moduleauthor:: Mark A. Conway-Greenslade


"""
import base64
import json
import os

import requests



# GitHub API - credentials.
_GH_API_CREDENTIALS = ('esdoc-system-user', os.getenv('ESDOC_GITHUB_ACCESS_TOKEN'))

# GitHub API - teams.
_GH_API_TEAMS = "https://api.github.com/teams"

# GitHub API - user.
_GH_API_USER = "https://api.github.com/user"

# Map of recognized teams and their GitHub identifiers.
_GH_TEAMS = {
    'cdf2cim-publication': 2375689,
    'documentation-publication': 2375693,
    'errata-publication': 2375691,

    'cmip5-bcc': 2567241,
    'cmip5-bnu': 2567242,
    'cmip5-cccma': 2567243,
    'cmip5-cmcc': 2567244,
    'cmip5-cnrm-cerfacs': 2567245,
    'cmip5-csiro': 2567246,
    'cmip5-csiro-bom': 2567247,
    'cmip5-csiro-qccce': 2567248,
    'cmip5-doe-cola-cmmap-gmu': 2567249,
    'cmip5-fio': 2567250,
    'cmip5-ichec': 2567251,
    'cmip5-inm': 2567252,
    'cmip5-ipsl': 2567253,
    'cmip5-lasg-cess': 2567254,
    'cmip5-lasg-iap': 2567255,
    'cmip5-miroc': 2567256,
    'cmip5-mohc': 2567257,
    'cmip5-mpi-m': 2567258,
    'cmip5-mri': 2567259,
    'cmip5-nasa-giss': 2567260,
    'cmip5-nasa-gmao': 2567261,
    'cmip5-ncar': 2567262,
    'cmip5-ncc': 2567263,
    'cmip5-nicam': 2567264,
    'cmip5-nimr-kma': 2567265,
    'cmip5-noaa-gfdl': 2567266,
    'cmip5-nsf-doe-ncar': 2567267,
    'cmip5-unsw': 2567268,
    'cmip6-awi': 2567269,
    'cmip6-bcc': 2567270,
    'cmip6-bnu': 2567271,
    'cmip6-cams': 2567272,
    'cmip6-cas': 2567273,
    'cmip6-cccma': 2567274,
    'cmip6-cccr-iitm': 2567275,
    'cmip6-cmcc': 2567276,
    'cmip6-cnrm-cerfacs': 2567277,
    'cmip6-csir-csiro': 2567278,
    'cmip6-csiro-bom': 2567279,
    'cmip6-ec-earth-consortium': 2567280,
    'cmip6-fio-ronm': 2567281,
    'cmip6-hammoz-consortium': 2567282,
    'cmip6-inm': 2567283,
    'cmip6-inpe': 2567284,
    'cmip6-ipsl': 2567285,
    'cmip6-messy-consortium': 2567286,
    'cmip6-miroc': 2567287,
    'cmip6-mohc': 2567288,
    'cmip6-mpi-m': 2567289,
    'cmip6-mri': 2567290,
    'cmip6-nasa-giss': 2567291,
    'cmip6-ncar': 2567292,
    'cmip6-ncc': 2567293,
    'cmip6-nerc': 2567294,
    'cmip6-nims-kma': 2567295,
    'cmip6-niwa': 2567296,
    'cmip6-noaa-gfdl': 2567297,
    'cmip6-nuist': 2567298,
    'cmip6-pcmdi': 2567299,
    'cmip6-snu': 2567300,
    'cmip6-thu': 2567301,
    'cordex-auth-lhtee': 2567302,
    'cordex-auth-met': 2567303,
    'cordex-awi': 2567304,
    'cordex-bccr': 2567305,
    'cordex-cccma': 2567306,
    'cordex-chmi': 2567307,
    'cordex-clmcom': 2567308,
    'cordex-cnrm': 2567309,
    'cordex-crp-gl': 2567310,
    'cordex-cuni': 2567311,
    'cordex-dhmz': 2567312,
    'cordex-dmi': 2567313,
    'cordex-enea': 2567314,
    'cordex-gerics': 2567315,
    'cordex-hms': 2567316,
    'cordex-ictp': 2567317,
    'cordex-idl': 2567318,
    'cordex-iitm': 2567319,
    'cordex-ipsl-ineris': 2567320,
    'cordex-knmi': 2567321,
    'cordex-mgo': 2567322,
    'cordex-miub': 2567323,
    'cordex-mohc': 2567324,
    'cordex-mpi-csc': 2567325,
    'cordex-nuim': 2567326,
    'cordex-rmib-ugent': 2567327,
    'cordex-smhi': 2567328,
    'cordex-ucan': 2567329,
    'cordex-uclm': 2567330,
    'cordex-uhoh': 2567331,
    'cordex-ulg': 2567332,
    'cordex-uqam': 2567333
}

# Request authentication error HTTP response code.
_HTTP_UNAUTHENTICATED_ERROR = 401

# Request authorization error HTTP response code.
_HTTP_UNAUTHORIZED_ERROR = 403


class AuthenticationError(Exception):
    """Raised when an authentication assertion fails.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(AuthenticationError, self).__init__("AUTHENTICATION FAILED")
        self.response_code = _HTTP_UNAUTHENTICATED_ERROR


class AuthorizationError(Exception):
    """Raised when an authorization assertion fails.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(AuthorizationError, self).__init__("AUTHORIZATION FAILED")
        self.response_code = _HTTP_UNAUTHORIZED_ERROR


def authenticate_user(credentials):
    """Authenticates user credentials request against GitHub user api.

    :param tuple credentials: 2 member tuple (GitHub username, GitHub access token)

    :returns: GitHub username
    :rtype: str

    """
    # Unpack credentials.
    user_id, _ = credentials

    # Invoke GitHub API.
    r = requests.get(_GH_API_USER, auth=credentials)

    # Assert access token is valid (status_code = 200).
    if r.status_code != 200:
        raise AuthenticationError()

    # Assert user has granted application read:org permissions.
    if 'read:org' not in r.headers['X-OAuth-Scopes']:
        raise AuthenticationError()

    # Assert user id matches access token.
    if json.loads(r.text)['login'] != user_id:
        raise AuthenticationError()

    return user_id


def authorize_user(team_id, user_id):
    """Verifies a user is a member of a team.

    :param str team_id: GitHub team identifier.
    :param str user_id: GitHub user login.

    """
    # Validate inputs.
    assert team_id in _GH_TEAMS, "Invalid team identifier {}".format(team_id)

    # Invoke GitHub API.
    url = '{}/{}/memberships/{}'.format(_GH_API_TEAMS, _GH_TEAMS[team_id], user_id)
    r = requests.get(url, auth=_GH_API_CREDENTIALS)

    # Assert user is a team member.
    if r.status_code != 200:
        raise AuthorizationError()


def strip_credentials(credentials):
    """Strips passed credentials from HTTP header.

    :param str credentials: Base64 encoded, ':' delimited credentials.

    :returns: Stripped credentials.
    :rtype: tuple

    """
    # Strip out HTTP Basic authorization header prefix.
    credentials = credentials.replace('Basic ', '')

    # Decode (b64).
    try:
        credentials = base64.b64decode(credentials)
    except TypeError:
        raise AuthenticationError()

    # Extract.
    credentials = credentials.split(':')
    if len(credentials) != 2:
        raise AuthenticationError()

    # Decode (utf-8).
    credentials = [i.decode('utf-8') for i in credentials]

    return tuple(credentials)



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean Timestepping Framework'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------

DETAILS['toplevel'] = {
    'description': 'Properties of time stepping in ocean',
    'properties': [
        ('overview','str','1.1',
            'Overview of time stepping in ocean'),
        ('diurnal_cycle', 'ENUM:diurnal_cycle_types', '1.1',
            'Diurnal cycle type'),
        ]
    }
# Tracers time stepping
DETAILS['toplevel:tracers'] = {
    'description': 'Properties of tracers time stepping in ocean',
    'properties': [
        ('scheme', 'ENUM:ocean_timestepping_types', '1.1',
            'Tracers time stepping scheme'),
        ('time_step', 'int', '1.1',
            'Tracers time step (in seconds)'),
        ]
    }
# Baroclinic dynamics
DETAILS['toplevel:baroclinic_dynamics'] = {
    'description': 'Baroclinic dynamics in ocean',
    'properties': [
        ('type', 'ENUM:baroclinic_types', '1.1',
            'Baroclinic dynamics type'),
        ('scheme', 'ENUM:ocean_timestepping_types', '1.1',
            'Baroclinic dynamics scheme'),
        ('time_step', 'int', '0.1',
            'Baroclinic time step (in seconds)'),
        ]
    }
# Barotropic
DETAILS['toplevel:barotropic'] = {
    'description': 'Barotropic time stepping in ocean',
    'properties': [
        ('splitting', 'ENUM:barotropic_splitting_methods', '1.1',
            'Time splitting method'),
        ('time_step', 'int', '0.1',
            'Barotropic time step (in seconds)'),
        ]
    }

# Vertical
DETAILS['toplevel:vertical_physics'] = {
    'description': 'Vertical physics time stepping in ocean',
    'properties': [
        ('method', 'str', '1.1',
            'Details of vertical time stepping in ocean'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['diurnal_cycle_types'] = {
    'description': 'Types of diurnal cycle resolution in ocean',
    'is_open': True,
    'members': [
        ('None', 'No diurnal cycle in ocean'),
        ('Via coupling', 'Diurnal cycle via coupling frequency'),
        ('Specific treatment', 'Specific treament'),
        ]
    }

ENUMERATIONS['ocean_timestepping_types'] = {
    'description': 'Type of timestepping scheme in ocean',
    'is_open': True,
    'members': [
        ('Leap-frog + Asselin filter', 'Leap-frog scheme with Asselin filter'),
        ('Leap-frog + Periodic Euler', 'Leap-frog scheme with Periodic Euler'),
        ('Predictor-corrector', 'Predictor-corrector scheme'),
        ('Runge-Kutta 2','Runge-Kutta 2 scheme'),
        ('AM3-LF', 'AM3-LF such as used in ROMS'),
        ('Forward-backward', 'Forward-backward scheme'),
        ('Forward operator', 'Forward operator scheme')
        ]
    }

ENUMERATIONS['baroclinic_types'] = {
    'description': 'Type of baroclinic dynamics in ocean',
    'is_open': True,
    'members': [
        ('Preconditioned conjugate gradient', None),
        ('Sub cyling', 'Sub cycling relative to tracers'),
        ]
    }

ENUMERATIONS['barotropic_splitting_methods'] = {
    'description': 'Methods for time splitting in ocean',
    'is_open': True,
    'members': [
        ('None', None),
        ('split explicit', None),
        ('implicit', None),
        ]
    }



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol model'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level aerosol model properties',
    'properties': [
        ('overview', 'str', '1.1',
            'Overview of atmosperic aerosol model'),
        ('processes', 'ENUM:model_processes', '1.N',
            'Processes included in the Aerosol model.'),
        ('coupling', 'ENUM:coupling', '0.N',
            'Other model components coupled to the Aerosol model'),
        ('gas_phase_precursors', 'ENUM:gas_phase_precursors', '1.N',
            'List of gas phase aerosol precursors.'),
        ('scheme_type', 'ENUM:scheme_types', '1.N',
             'Type(s) of aerosol scheme used by the aerosols model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).'),
        ('bulk_scheme_species', 'ENUM:species', '1.N',
             'List of species covered by the bulk scheme.'),
    ]
}

# --------------------------------------------------------------------
# PROCESS: Optical and radiative properties
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties'] = {
    'description': 'Optical and radiative properties',
    'properties': [
        ('overview', 'str', '1.1',
             'Overview of optical and radiative properties'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Absorption
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties:absorption'] = {
    'description': 'Absortion properties in aerosol scheme',
    'properties': [
        ('black_carbon', 'float', '0.1',
             'Absorption mass coefficient of black carbon at 550nm'),
        ('dust', 'float', '0.1',
             'Absorption mass coefficient of dust at 550nm'),
        ('organics', 'float', '0.1',
             'Absorption mass coefficient of organics at 550nm'),
        ('non_absorbing_state', 'bool', '0.1',
             'Do organics have non-absorbing state "Phi"?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Mixtures
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties:mixtures'] = {
    'description': '',
    'properties': [
        ('external', 'bool', '1.1',
            'Is there external mixing with respect to chemical composition?'),
        ('internal', 'bool', '1.1',
            'Is there internal mixing with respect to chemical composition?'),
        ('mixing_rule', 'str', '0.1',
             'If there is internal mixing with respect to chemical composition then indicate the mixinrg rule'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Impact of H2O
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties:impact_of_h2o'] = {
    'description': '',
    'properties': [
        ('size', 'bool', '1.1',
            'Does H2O impact size?'),
        ('internal_mixture', 'bool', '1.1',
            'Does H2O impact internal mixture?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Radiative scheme for aerosol
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties:radiative_scheme'] = {
    'description': 'Radiative scheme for aerosol',
    'properties': [
        ('overview', 'str', '1.1',
             'Overview of radiative scheme'),      
        ('shortwave_bands', 'int', '1.1',
            'Number of shortwave bands'),
        ('longwave_bands', 'int', '1.1',
            'Number of longwave bands'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Radiative scheme for aerosol
# --------------------------------------------------------------------
DETAILS['optical_radiative_properties:cloud_interactions'] = {
    'description': 'Aerosol-cloud interactions',
    'properties': [
        ('overview', 'str', '1.1',
             'Overview of aerosol-cloud interactions'),
        ('twomey', 'bool', '1.1',
            'Is the Twomey effect included?'),
        ('twomey_minimum_ccn', 'int', '0.1',
            'If the Twomey effect is included, then what is the minimum CCN number?'),
        ('drizzle', 'bool', '1.1',
            'Does the scheme affect drizzle?'),
        ('cloud_lifetime', 'bool', '1.1',
            'Does the scheme affect cloud lifetime?'),
        ('longwave_bands', 'int', '1.1',
            'Number of longwave bands'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_processes'] = {
    'description': 'Processes included in the Aerosol model.',
    'is_open': False,
    'members': [
        ('Dry deposition', None),
        ('Sedimentation', None),
        ('Wet deposition (impaction scavenging)', None),
        ('Wet deposition (nucleation scavenging)', None),
        ('Coagulation', None),
        ('Oxidation (gas phase)', None),
        ('Oxidation (in cloud)', None),
        ('Condensation', None),
        ('Ageing', None),
        ('Advection (horizontal)', None),
        ('Advection (vertical)', None),
        ('Heterogeneous chemistry', None),
        ('Nucleation', None),
    ]
}

ENUMERATIONS['coupling'] = {
    'description':'Other model components coupled to the Aerosol model.',
    'is_open': True,
    'members':[
        ('Radiation', None),
        ('Land surface', None),
        ('Heterogeneous chemistry', None),
        ('Clouds', None),
        ('Ocean', None),
        ('Cryosphere', None),
        ('Gas phase chemistry', None),
    ]
}

ENUMERATIONS['gas_phase_precursors'] = {
    'description': 'List of gas phase aerosol precursors.',
    'is_open': True,
    'members':[
        ('DMS', None),
        ('SO2', None),
        ('Ammonia', None),
        ('Iodine', None),
        ('Terpene', None),
        ('Isoprene', None),
        ('VOC', None),
        ('NOx', None),
        
    ]
}

ENUMERATIONS['scheme_types'] = {
    'description': 'Type(s) of aerosol scheme used by the aerosols model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).',
    'is_open': True,
    'members':[
        ('Bulk', None),
        ('Modal', None),
        ('Bin', None),
    ]
}

ENUMERATIONS['species'] = {
    'description':'List of species.',
    'is_open': True,
    'members':[
        ('Sulphate', None),
        ('Nitrate', None),
        ('Sea salt', None),
        ('Dust', None),
        ('Ice', None),
        ('Organic', None),
        ('Black carbon / soot', None),
        ('SOA (secondary organic aerosols)', None),
        ('POM (particulate organic matter)', None),
        ('Polar stratospheric ice', None),
        ('NAT (Nitric acid trihydrate)', None),
        ('NAD (Nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particule)', None),
    ]
}




"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Key properties of the atmospheric chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS = OrderedDict()

DETAILS['toplevel'] = {
    'description': 'Top level key properties in atmospheric chemistry model',
    'properties': [
        ('model_overview', 'str', '1.1',
             'Overview of atmospheric chemistry model.'),
        ('model_name', 'str', '1.1',
             'Name of atmospheric chemistry model code'),
        ('chemistry_scheme_scope', 'ENUM:chemistry_scheme_scopes', '1.N',
            'Atmospheric domains covered by the atmospheric chemistry model'),
        ('basic_approximations', 'str', '1.1',
            'Basic approximations made in the atmospheric chemistry model',),
        ('prognostic_variables_form', 'ENUM:prognostic_vars_types', '1.N',
            'Form of prognostic variables in the atmospheric chemistry component.'),
        ('number_of_tracers', 'int', '1.1',
            'Number of advected tracers in the atmospheric chemistry model'),
        ('family_approach', 'bool', '1.1',
            'Atmospheric chemistry calculations (not advection) generalized into families of species?'),
        ('coupling_with_chemical_reactivity', 'bool', '1.1',
            'Atmospheric chemistry transport scheme turbulence is couple with chemical reactivity?'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: timestepping framework
# --------------------------------------------------------------------
DETAILS['timestep_framework'] = {
    'description': 'Timestepping in the atmospheric chemistry model',
    'properties' : [
        ('method', 'ENUM:timestepping_methods', '1.1',
            'Mathematical method deployed to solve the evolution of a given variable'),
        ('split_operator_advection_timestep', 'int', '0.1',
            'Timestep for chemical species advection (in seconds)'),
        ('split_operator_physical_timestep', 'int', '0.1',
            'Timestep for physics (in seconds).'),
        ('split_operator_chemistry_timestep', 'int', '0.1',
            'Timestep for chemistry (in seconds).'),
        ('split_operator_alternate_order', 'bool', '0.1',
            '?'),
        ('integrated_timestep', 'int', '1.1',
            'Timestep for the atmospheric chemistry model (in seconds)'),
        ('integrated_scheme_type', 'ENUM:integrated_scheme_types', '1.1',
            'Specify the type of timestep scheme'),
    ],
    'detail_sets': [
        'split_operator_order',
    ]
}

_note1 = ' This should be an integer greater than zero, and may be the same value as for another process if they are calculated at the same time.'


DETAILS['timestep_framework:split_operator_order'] = {
    'description': '',
    'properties' : [
        ('turbulence', 'int', '0.1',
            'Call order for turbulence scheme.'+_note1),
        ('convection', 'int', '0.1',
            'Call order for convection scheme'+_note1),
        ('precipitation', 'int', '0.1',
            'Call order for precipitation scheme.'+_note1),
        ('emissions', 'int', '0.1',
            'Call order for emissions scheme.'+_note1),
        ('deposition', 'int', '0.1',
            'Call order for deposition scheme.'+_note1),
        ('gas_phase_chemistry', 'int', '0.1',
            'Call order for gas phase chemistry scheme.'+_note1),
        ('tropospheric_heterogeneous_phase_chemistry', 'int', '0.1',
            'Call order for tropospheric heterogeneous phase chemistry scheme.'+_note1),
        ('stratospheric_heterogeneous_phase_chemistry', 'int', '0.1',
            'Call order for stratospheric heterogeneous phase chemistry scheme.'+_note1),
        ('photo_chemistry', 'int', '0.1',
            'Call order for photo chemistry scheme.'+_note1),
        ('aerosols', 'int', '0.1',
            'Call order for aerosols scheme.'+_note1),
    ]
}

## --------------------------------------------------------------------
## SUB-PROCESS: Meteorological forcings
## --------------------------------------------------------------------
#DETAILS['meteorological_forcings'] = {
#    'description': '',
#    'properties' : [
#        ('variables_3D', 'str', '0.1',
#            'Three dimensionsal forcing variables, e.g. U, V, W, T, Q, P, conventive mass flux'),
#        ('variables_2D', 'str', '0.1',
#            'Two dimensionsal forcing variables, e.g. land-sea mask definition'),
#        ('frequency', 'int', '1.1',
#            'Frequency with which meteological forcings are applied (in seconds).'),
#        ]
#}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the atmospheric chemistry grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '0.1',
             "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '0.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '0.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '0.1',
             "Default is False. Set true if grid resolution changes during execution."),
        ],
    }

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for atmospheric chemistry component',
    'properties': [
        ('description', 'str', '1.1',
             "General overview description of tuning: explain and motivate the main targets and metrics retained. &"
             "Document the relative weight given to climate performance metrics versus process oriented metrics, &"
             "and on the possible conflicts with parameterization level tuning. In particular describe any struggle &"
             "with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'str', '0.N',
             "List set of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'str', '0.N',
             "List of regional metrics of mean state used in tuning model/component"),
        ('trend_metrics_used', 'str', '0.N',
             "List observed trend metrics used in tuning model/component"),
        ]
    }

DETAILS['toplevel:software_properties'] = {
    'description': 'Software properties of aerosol code',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','str', '0.N',
            "Code language(s)."),
    ]
}

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS = OrderedDict()

ENUMERATIONS['chemistry_scheme_scopes'] = {
    'description': 'Atmospheric domains covered by the atmospheric chemistry model',
    'is_open': True,
    'members': [
        ('troposhere', None),
        ('stratosphere', None),
        ('mesosphere', None),
        ('mesosphere', None),
        ('whole atmosphere', None),
    ]
}

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'Form of prognostic variables in the atmospheric chemistry component.',
    'is_open': True,
    'members': [
        ('3D mass/mixing ratio for gas', None),
    ]
}

ENUMERATIONS['timestepping_methods'] = {
    'description': 'Mathematical method deployed to solve the evolution of a given variable',
    'is_open': True,
    'members': [
        ('Operator splitting', None),
        ('Integrated', None),
    ]
}
     
ENUMERATIONS['integrated_scheme_types'] = {
    'description': 'Specify the type of timestep scheme',
    'is_open': True,
    'members': [
        ('Explicit', None),
        ('Implicit', None),
        ('Semi-implicit', None),
        ('Semi-analytic', None),
        ('Impact solver', None),
        ('Back Euler', None),
        ('Newton Raphson', None),
        ('Rosenbrock', None),
    ]
}



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Large Scale Cloud Microphysics and Precipitation'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------

DETAILS['toplevel'] = {
    'description': "Overview of large scale cloud microphysics and precipitation",
    'properties': [
        ('overview', 'str', '1.1',
            'Overview description of large scale cloud microphysics and precipitation'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: large_scale_precipitation
# --------------------------------------------------------------------
DETAILS['large_scale_precipitation'] = {
    'description': 'Properties of the large scale precipitation scheme',
    'properties': [
        ('scheme_name', 'str', '0.1',
            'Commonly used name of the large scale precipitation parameterisation scheme'),
        ('hydrometeors', 'ENUM:hydrometeor_types', '1.N',
            'Precipitating hydrometeors taken into account in the large scale precipitation scheme'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: large_scale_cloud_microphysics
# --------------------------------------------------------------------
DETAILS['large_scale_cloud_microphysics'] = {
    'description': 'Properties of the large scale cloud microphysics scheme',
    'properties': [
        ('scheme_name', 'str', '0.1',
            'Commonly used name of the microphysics parameterisation scheme used for large scale clouds.'),
        ('processes', 'ENUM:processes_attributes', '1.N',
            'Large scale cloud microphysics processes'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['hydrometeor_types'] = {
    'description': 'Precipitating hydrometeors taken into account in the large scale precipitation scheme',
    'is_open': True,
    'members': [
        ('liquid rain', None),
        ('snow', None),
        ('hail', None),
        ('graupel', None),
        ]
    }

ENUMERATIONS['processes_attributes'] = {
    'description': 'Cloud microphysics processes',
    'is_open': True,
    'members': [
        ('mixed phase', None),
        ('cloud droplets', None),
        ('cloud ice', None),
        ('ice nucleation', None),
        ('water vapour deposition', None),
        ('effect of raindrops', None),
        ('effect of snow', None),
        ('effect of graupel', None),
        ]
    }




"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry tropospheric heterogeneous chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Tropospheric heterogenous atmospheric chemistry',
    'properties': [ 
        ('overview', 'str', '1.1',
             'Overview tropospheric heterogenous atmospheric chemistry'),
        ('gas_phase_species', 'str', '0.1',
            'List of gas phase species included in the tropospheric heterogeneous chemistry scheme.'),
        ('aerosol_species', 'ENUM:aerosol_species', '0.N',
            'Aerosol species included in the tropospheric heterogeneous chemistry scheme.'),
        ('number_of_steady_state_species', 'int', '1.1',
             'The number of steady state species in the tropospheric heterogeneous chemistry scheme.'),
        ('interactive_dry_deposition', 'bool', '1.1',
            'Is dry deposition interactive (as opposed to prescribed)? Dry deposition describes the dry processes by which gaseous species deposit themselves on solid surfaces thus decreasing their concentration in the air.'),  
        ('coagulation', 'bool', '1.1',
            'Is coagulation is included in the tropospheric heterogeneous chemistry scheme or not?'),       
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['aerosol_species'] = {
    'description': 'Aerosol species included in the stratospheric heterogeneous chemistry scheme',
    'is_open': False,
    'members': [
        ('Sulphate', None),
        ('Nitrate', None),
        ('Sea salt', None),
        ('Dust', None),
        ('Ice', None),
        ('Organic', None),
        ('Black carbon/soot', None),
        ('Polar stratospheric ice', None),
        ('Secondary organic aerosols', None),
        ('Particulate organic matter', None),
    ]
}
